'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _backo = require('backo');

var _backo2 = _interopRequireDefault(_backo);

var _invariant = require('invariant');

var _invariant2 = _interopRequireDefault(_invariant);

var _lodash = require('lodash.identity');

var _lodash2 = _interopRequireDefault(_lodash);

var _lodash3 = require('lodash.includes');

var _lodash4 = _interopRequireDefault(_lodash3);

var _superagent = require('superagent');

var _superagent2 = _interopRequireDefault(_superagent);

var _actions = require('../actions');

var _actionTypes = require('../constants/action-types');

var actionTypes = _interopRequireWildcard(_actionTypes);

var _httpMethods = require('../constants/http-methods');

var httpMethods = _interopRequireWildcard(_httpMethods);

var _statusCodes = require('../constants/status-codes');

var statusCodes = _interopRequireWildcard(_statusCodes);

var _queryKey3 = require('../lib/query-key');

var _immutable = require('immutable');

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var createRequest = function createRequest(url, method) {
    var request = void 0;
    switch (method) {
        case httpMethods.GET:
            request = _superagent2.default.get(url);
            break;
        case httpMethods.POST:
            request = _superagent2.default.post(url);
            break;
        case httpMethods.PUT:
            request = _superagent2.default.put(url);
            break;
        case httpMethods.DELETE:
            request = _superagent2.default.del(url);
            break;
        default:
            throw new Error('Unsupported HTTP method: ' + method);
    }

    return request;
};

var updateEntities = function updateEntities(update) {
    var entities = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : (0, _immutable.Map)();
    var transformed = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : (0, _immutable.Map)();

    // If update, not supplied, then no change to entities should be made
    return Object.keys(update || {}).reduce(function (accum, key) {
        return accum.set(key, update[key](entities.get(key), transformed.get(key)));
    }, new _immutable.Map());
};

var optimisticUpdateEntities = function optimisticUpdateEntities(optimisticUpdate) {
    var entities = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : (0, _immutable.Map)();

    return Object.keys(optimisticUpdate).reduce(function (accum, key) {
        if (optimisticUpdate[key]) {
            return accum.set(key, optimisticUpdate[key](entities.get(key)));
        }
        return accum.set(key, entities.get(key));
    }, new _immutable.Map());
};

var defaultConfig = {
    backoff: {
        maxAttempts: 5,
        minDuration: 300,
        maxDuration: 5000
    },
    retryableStatusCodes: [statusCodes.UNKNOWN, // normally means a failed connection
    statusCodes.REQUEST_TIMEOUT, statusCodes.TOO_MANY_REQUESTS, // hopefully backoff stops this getting worse
    statusCodes.SERVICE_UNAVAILABLE, statusCodes.GATEWAY_TIMEOUT]
};

var getPendingQueries = function getPendingQueries(queries) {
    return queries.filter(function (query) {
        return query.get('isPending');
    });
};

var queryMiddleware = function queryMiddleware(queriesSelector, entitiesSelector) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : defaultConfig;

    return function (_ref) {
        var dispatch = _ref.dispatch,
            getState = _ref.getState;
        return function (next) {
            return function (action) {
                // TODO(ryan): add warnings when there are simultaneous requests and mutation queries for the same entities
                var returnValue = void 0;

                switch (action.type) {
                    case actionTypes.REQUEST_ASYNC:
                        {
                            var url = action.url,
                                body = action.body,
                                force = action.force,
                                retry = action.retry,
                                _action$transform = action.transform,
                                transform = _action$transform === undefined ? _lodash2.default : _action$transform,
                                update = action.update,
                                _action$options = action.options,
                                options = _action$options === undefined ? {} : _action$options,
                                meta = action.meta;


                            (0, _invariant2.default)(!!url, 'Missing required `url` field in action handler');
                            (0, _invariant2.default)(!!update, 'Missing required `update` field in action handler');

                            var queryKey = (0, _queryKey3.reconcileQueryKey)(action);

                            var state = getState();
                            var queries = queriesSelector(state);
                            var queriesState = queries.get(queryKey, new _immutable.Map());
                            var isPending = queriesState.get('isPending');
                            var status = queriesState.get('status');
                            var hasSucceeded = status >= 200 && status < 300;

                            if (force || queriesState.isEmpty() || retry && !isPending && !hasSucceeded) {
                                returnValue = new Promise(function (resolve) {
                                    var start = new Date();
                                    var _options$method = options.method,
                                        method = _options$method === undefined ? httpMethods.GET : _options$method;


                                    var request = createRequest(url, method);

                                    if (body) {
                                        request.send(body);
                                    }

                                    if (options.headers) {
                                        request.set(options.headers);
                                    }

                                    if (options.credentials === 'include') {
                                        request.withCredentials();
                                    }

                                    var attempts = 0;
                                    var backoff = new _backo2.default({
                                        min: config.backoff.minDuration,
                                        max: config.backoff.maxDuration
                                    });

                                    var attemptRequest = function attemptRequest() {
                                        dispatch((0, _actions.requestStart)(url, body, request, meta, queryKey));
                                        attempts += 1;
                                        request.end(function (err, response) {
                                            var resOk = !!(response && response.ok);
                                            var resStatus = response && response.status || 0;
                                            var resBody = response && response.body || undefined;
                                            var resText = response && response.text || undefined;

                                            var transformed = void 0;
                                            var newEntities = void 0;

                                            if ((0, _lodash4.default)(config.retryableStatusCodes, resStatus) && attempts < config.backoff.maxAttempts) {
                                                // TODO take into account Retry-After header if 503
                                                setTimeout(attemptRequest, backoff.duration());
                                                return;
                                            }

                                            if (err || !resOk) {
                                                dispatch((0, _actions.requestFailure)(url, body, resStatus, resBody, meta, queryKey));
                                            } else {
                                                var callbackState = getState();
                                                var entities = entitiesSelector(callbackState);
                                                transformed = (0, _immutable.fromJS)(transform(resBody, resText));
                                                newEntities = updateEntities(update, entities, transformed);
                                                dispatch((0, _actions.requestSuccess)(url, body, resStatus, newEntities, meta, queryKey));
                                            }

                                            var end = new Date();
                                            var duration = end - start;
                                            resolve({
                                                body: resBody,
                                                duration: duration,
                                                status: resStatus,
                                                text: resText,
                                                transformed: transformed,
                                                entities: newEntities
                                            });
                                        });
                                    };

                                    attemptRequest();
                                });
                            }

                            break;
                        }
                    case actionTypes.MUTATE_ASYNC:
                        {
                            var _url = action.url,
                                _action$transform2 = action.transform,
                                _transform = _action$transform2 === undefined ? _lodash2.default : _action$transform2,
                                _update = action.update,
                                _body = action.body,
                                optimisticUpdate = action.optimisticUpdate,
                                _action$options2 = action.options,
                                _options = _action$options2 === undefined ? {} : _action$options2;

                            (0, _invariant2.default)(!!_url, 'Missing required `url` field in action handler');

                            var _state = getState();
                            var entities = entitiesSelector(_state);
                            var optimisticEntities = void 0;
                            if (optimisticUpdate) {
                                optimisticEntities = optimisticUpdateEntities(optimisticUpdate, entities);
                            }

                            var _queryKey = (0, _queryKey3.reconcileQueryKey)(action);

                            returnValue = new Promise(function (resolve) {
                                var start = new Date();
                                var _options$method2 = _options.method,
                                    method = _options$method2 === undefined ? httpMethods.POST : _options$method2;


                                var request = createRequest(_url, method);

                                if (_options.headers) {
                                    request.set(_options.headers);
                                }

                                if (_options.credentials === 'include') {
                                    request.withCredentials();
                                }
                                // Note: only the entities that are included in `optimisticUpdate` will be passed along in the
                                // `mutateStart` action as `optimisticEntities`
                                dispatch((0, _actions.mutateStart)(_url, _body, request, optimisticEntities, _queryKey));

                                request.send(_body).end(function (err, response) {
                                    var resOk = !!(response && response.ok);
                                    var resStatus = response && response.status || 0;
                                    var resBody = response && response.body || undefined;
                                    var resText = response && response.text || undefined;

                                    var transformed = void 0;
                                    var newEntities = void 0;

                                    if (err || !resOk) {
                                        dispatch((0, _actions.mutateFailure)(_url, _body, resStatus, entities, _queryKey));
                                    } else {
                                        transformed = (0, _immutable.fromJS)(_transform(resBody, resText));
                                        newEntities = updateEntities(_update, entities, transformed);
                                        dispatch((0, _actions.mutateSuccess)(_url, _body, resStatus, newEntities, _queryKey));
                                    }

                                    var end = new Date();
                                    var duration = end - start;
                                    resolve({
                                        body: resBody,
                                        duration: duration,
                                        status: resStatus,
                                        text: resText,
                                        transformed: transformed,
                                        entities: newEntities
                                    });
                                });
                            });

                            break;
                        }
                    case actionTypes.CANCEL_QUERY:
                        {
                            var _queryKey2 = action.queryKey;

                            (0, _invariant2.default)(!!_queryKey2, 'Missing required `queryKey` field in action handler');

                            var _state2 = getState();
                            var _queries = queriesSelector(_state2);
                            var pendingQueries = getPendingQueries(_queries);

                            if (pendingQueries.has(_queryKey2)) {
                                pendingQueries.getIn([_queryKey2, 'request', 'abort'])();
                                returnValue = next(action);
                            } else {
                                console.warn('Trying to cancel a request that is not in flight: ', _queryKey2);
                                returnValue = null;
                            }

                            break;
                        }
                    case actionTypes.RESET:
                        {
                            var _state3 = getState();
                            var _queries2 = queriesSelector(_state3);
                            var _pendingQueries = getPendingQueries(_queries2);

                            _pendingQueries.forEach(function (query) {
                                return query.getIn(['request', 'abort'])();
                            });
                            returnValue = next(action);

                            break;
                        }
                    default:
                        {
                            returnValue = next(action);
                        }
                }

                return returnValue;
            };
        };
    };
};

exports.default = queryMiddleware;